@@ -7,32 +7,65 @@
  'bin/zopectl stop' saves on the number of terminals you have to keep
  open.
 
- The 'ACTIONS' dictionary at the end of this file is the list of
- user-visible actions. There's no need for every action to be listed
- there, so you can define Actions just for internal use, if you want.
 """
 
+import actionutils
 import config
-import sys
+import logging
 import os
 import os.path
-import logging
 import shutil
+import sys
+import tempfile
+from actionutils import InvokedInVolatileDir
 log = logging.getLogger('actions')
 
-
 class Action:
     """Base class for all actions.
+
+    name -- The short name used in the optionparser's help message.
+
+    option -- Option string like '--option', if not present (like in
+    this base class), it will not be turned into an option.
     """
 
     name = 'abstract base action'
+    option = None
+    alternativeOption = None
+    kindOfAction = 'store_true'
 
     def __init__(self, configuration):
         self.conf = configuration
 
-    def run(self):
+    def run(self, options=None):
         pass
 
+    def addOption(self, optionparser):
+        optionparser.add_option(
+            self.option,
+            self.alternativeOption,
+            help=self.name,
+            action=self.kindOfAction,
+            default=False)
+
+class ArgumentAction(Action):
+    """Base action that accepts one parameter.
+
+    The normal actions are only switched on (default is off), this
+    action accepts additional parameters.
+    """
+
+    kindOfAction = 'store' #TODO
+    metavar = 'CMD'
+
+    def addOption(self, optionparser):
+        optionparser.add_option(
+            self.option,
+            self.alternativeOption,
+            help=self.name,
+            metavar=self.metavar,
+            action=self.kindOfAction)
+
 
 class CreateInstanceAction(Action):
     """Create a zope instance, possibly including a zeo server.
@@ -48,19 +81,23 @@ class CreateInstanceAction(Action):
     """
 
     name = "Create a zope instance for your project."
+    option = '--create'
 
-    def run(self):
+    def run(self, options=None):
         useZeo = self.conf.configData['use_zeo']
         python = self.conf.configData['python']
         instanceDir = self.conf.instanceDir()
         zeoDir = self.conf.zeoDir()
-        for dir in [instanceDir, zeoDir]:
+        dirs_to_remove = [instanceDir]
+        if useZeo:
+            dirs_to_remove.append(zeoDir)
+        for dir in dirs_to_remove:
+            try:
+                actionutils.trapUsageWithinVolatileDir(dir)
+            except InvokedInVolatileDir, error:
+                log.error(error)
+                sys.exit(1)
             if os.path.exists(dir):
-                if os.getcwd().startswith(dir):
-                    log.error("Current working directory is in instance, "
-                              "which would cause an error while making "
-                              "a Zope/Zeo instance.")
-                    sys.exit(1)
                 shutil.rmtree(dir)
                 log.debug("Removed instance/zeo dir '%s'.", dir)
         # Zope server
@@ -176,8 +213,9 @@ class CopyDatafsAction(Action):
     """
 
     name = "Copy over a fresh, prepared, 'Data.fs'."
+    option = '--copydatafs'
 
-    def run(self):
+    def run(self, options=None):
         source = self.conf.datafs()
         # databasepath keeps track of zeo usage
         target = self.conf.databasePath()
@@ -190,7 +228,6 @@ class CopyDatafsAction(Action):
             log.info("Copied over fresh Data.fs from %s.", source)
             log.debug("Copied it to %s.", target)
 
-
 class AddProductsAction(Action):
     """Rebuild the Products/ directory from scratch.
 
@@ -205,9 +242,15 @@ class AddProductsAction(Action):
     """
 
     name = "Rebuild the Products/ directory ."
+    option = '--products'
 
-    def run(self):
+    def run(self, options=None):
         target = os.path.join(self.conf.instanceDir(), 'Products')
+        try:
+            actionutils.trapUsageWithinVolatileDir(target)
+        except InvokedInVolatileDir, error:
+            log.error(error)
+            sys.exit(1)
         shutil.rmtree(target)
         log.debug("Removed the entire %s directory.", target)
         os.mkdir(target)
@@ -219,198 +262,35 @@ class AddProductsAction(Action):
             except IOError:
                 log.warn('Please fix this.')
                 # XXX Maybe we want to quit with an error here.
-        log.info("Rebuild the Products/ directory from scratch.")
+        log.info("Rebuilt the Products/ directory from scratch.")
 
 
-class FreshInstallAction(Action):
-    """Wipe everything and create an instance including products.
-
-    The "fresh" action is the most brute-force way to get a fresh
-    instance. It will stop zope (when running), remove the instance
-    completely, set up a new instance, copy the Data.fs (if you have
-    one prepared), fill your Products directory, quickreinstall your
-    products and start zope. Pfew!
-
-    This is, really, the **main action you should be using**, apart
-    from the "soft" action. It makes it easy to get to a known, clear,
-    fresh starting position without old cruft, locally modified files,
-    etc.
+class ZopectlAction(ArgumentAction):
+    """Action for calling zopectl.
     """
 
-    name = "Wipe everything and create an instance " +\
-        "including products."
+    name = "Runs your instance's 'bin/zopectl CMD'."
+    option = '--zope'
+    alternativeOption = '-z'
 
-    def run(self):
-        useZeo = self.conf.configData['use_zeo']
-        if useZeo:
-            subActions = [                          
-                StopZeoAction,
-                StopZopeAction,
-                #WaitAction,
-                CreateInstanceAction,
-                CopyDatafsAction,
-                StartZeoAction,
-                AddProductsAction,
-                StartZopeAction,
-                QuickreinstallAction,
-                ]
-        else:
-            subActions = [StopZopeAction,
-                          CreateInstanceAction,
-                          CopyDatafsAction,
-                          AddProductsAction,
-                          QuickreinstallAction,
-                          StartZopeAction,
-                          ]
-        for actionClass in subActions:
-            action = actionClass(self.conf)
-            action.run()
+    def run(self, options=None):
+        command = options.zope
+        actionutils.runZopectl(self.conf, command)
 
-        
-class SoftInstallAction(Action):
-    """Restart zope and call the quickreinstall script.
 
-    The "soft" action is the **second main action** you should be
-    using. It stops zope, quickreinstalls your products and starts
-    zope again.
-
-    In many development scenarios, zope will figure out you changed
-    something (like a page template) when running in debug mode. To
-    actually see the changes you made in python classes, you need to
-    restart zope ('instancemanager yourproject restart'). When working
-    on the install methods, for instance for modifying portal
-    settings, the only alternative is to restart and to click around
-    in the quickinstaller time after time. Too much work, so use this
-    action :-) 
+class ZeoctlAction(ArgumentAction):
+    """Action for calling zeoctl.
     """
 
-    name = "Restart zope and call the quickreinstall script."
+    name = "Runs your zeo server's 'bin/zeoctl CMD'."
+    option = '--zeo'
 
-    def run(self):
-        useZeo = self.conf.configData['use_zeo']
-        if useZeo:
-            subActions = [QuickreinstallAction]
-        else:
-            subActions = [StopZopeAction,
-                          QuickreinstallAction,
-                          StartZopeAction,
-                          ]
-        for actionClass in subActions:
-            action = actionClass(self.conf)
-            action.run()
-
-        
-class BaseZopeAction(Action):
-    """Base action for starting/stopping zope
-    """
-
-    name = "Base for the zope actions."
-    argument = 'start'
-
-    def run(self):
-        zopectlCommand = os.path.join(self.conf.instanceDir(), 
-                                      'bin',
-                                      'zopectl')
-        command = ' '.join([zopectlCommand, self.argument])
-        log.debug("Calling zope:")
-        log.debug(command)
-        if os.path.exists(zopectlCommand):
-            os.system(command)
-        else:
-            log.error("%s does not exist", zopectlCommand)
+    def run(self, options=None):
+        command = options.zeo
+        actionutils.runZeoctl(self.conf, command)
 
 
-class BaseZeoAction(Action):
-    """Base action for starting/stopping zeo
-    """
-
-    name = "Base for the zeo actions."
-    argument = 'start'
-
-    def run(self):
-        zeoctlCommand = os.path.join(self.conf.zeoDir(), 
-                                      'bin',
-                                      'zeoctl')
-        command = ' '.join([zeoctlCommand, self.argument])
-        log.debug("Calling zeo:")
-        log.debug(command)
-        if os.path.exists(zeoctlCommand):
-            os.system(command)
-        else:
-            log.error("%s does not exist", zeoctlCommand)
-
-
-class StartZeoAction(BaseZeoAction):
-    """Start the zeo instance.
-    """
-
-    name = "Start the zeo instance."
-    argument = 'start'
-    
-    def run(self):
-        BaseZeoAction.run(self)
-        log.info("Started zeo on port %s.",
-                 self.conf.zeoPort())
-
-
-class StopZeoAction(BaseZeoAction):
-    """Stop the zeo instance.
-    """
-
-    name = "Stop the zeo instance."
-    argument = 'stop'
-    
-    def run(self):
-        BaseZeoAction.run(self)
-        log.info("Stopped zeo on port %s.",
-                 self.conf.zeoPort())
-
-
-class StartZopeAction(BaseZopeAction):
-    """Start the zope instance.
-    """
-
-    name = "Start the zope instance."
-    argument = 'start'
-    
-    def run(self):
-        BaseZopeAction.run(self)
-        log.info("Started zope on port %s.",
-                 self.conf.configData['port'])
-
-
-class RestartZopeAction(BaseZopeAction):
-    """Restart the zope instance.
-    """
-
-    name = "Restart the zope instance."
-    argument = 'restart'
-
-    def run(self):
-        BaseZopeAction.run(self)
-        log.info("Restarted zope on port %s.",
-                 self.conf.configData['port'])
-
-
-class StopZopeAction(BaseZopeAction):
-    """Stop the zope instance.
-    """
-
-    name = "Stop the zope instance."
-    argument = 'stop'
-
-
-class ForegroundZopeAction(BaseZopeAction):
-    """Start the zope instance in the foreground.
-
-    When the zope instance is running already this does nothing.
-    """
-
-    name = "Foreground the zope instance."
-    argument = 'fg'
-
-
-class QuickreinstallAction(BaseZopeAction):
+class QuickreinstallAction(Action):
     """Action that quickreinstalls your products.
 
     This used to depend on the presence of a specific script in your
@@ -421,6 +301,8 @@ class QuickreinstallAction(BaseZopeAction):
     """
 
     name = "Action that quickreinstalls your products."
+    option = '--reinstall'
+    alternativeOption = '-r'
 
     def scriptLocation(self):
         """Return the location of the quickreinstall script.
@@ -431,7 +313,7 @@ class QuickreinstallAction(BaseZopeAction):
         log.debug("Quickinstall script location is %s.", location)
         return location
 
-    def run(self):
+    def run(self, options=None):
         ploneSiteName = self.conf.configData['plone_site_name']
         if not ploneSiteName:
             log.debug("No plone site name, so no reinstall.")
@@ -449,10 +331,8 @@ class QuickreinstallAction(BaseZopeAction):
         #arguments.append(u'profile-Products.mauritssetup:mauritssetup')
 
         # Now prepare it for the BaseZopeAction.
-        self.argument = ' '.join(arguments)
-        log.debug("Running zopectl with these arguments: %s.",
-                  self.argument)
-        BaseZopeAction.run(self)
+        command = ' '.join(arguments)
+        actionutils.runZopectl(self.conf, command)
         log.info("Quick reinstall script called.")
 
 
@@ -465,8 +345,9 @@ class PrintConfigAction(Action):
     """
 
     name = "Print the configuration for this project."
+    option = '--printconfig'
 
-    def run(self):
+    def run(self, options=None):
         configList = self.conf.configData.keys()
         configList.sort()
         for option in configList:
@@ -476,32 +357,98 @@ class BackupZopeAction(Action):
     """Backup the Zope Database of the instance.
 
     This action relies on the backup_basedir_template variable to store the
-    backups.
+    backups. By default, the backups are incremental unless there's
+    been a pack.
     """
 
-    name = "Backup the instance database."
-
-    def run(self):
+    name = "Backup the instance database (incremental backup)."
+    option = '--backup'
+    alternativeOption = '-b'
+    
+    def run(self, options=None):
         # read paths from configuration
-        backup_base_dir = self.conf.backupBaseDir()
-        target = os.path.join(backup_base_dir, self.conf.configData['project'])
+        target = self.conf.backupBaseDir()
         source = self.conf.databasePath()
+        actionutils.backup(conf=self.conf,
+                           sourceDatafs=source,
+                           targetDir=target,
+                           full=False)
 
-        # if the backup directory doesn't exist, create it.
-        if not os.path.exists(target):
-            os.makedirs(target)
-            log.info("%s does not exists, creating it now." % target)
-        arguments = "--backup --file=%s --repository=%s" % (source, target)
-        python = self.conf.configData['python']
-        repozo = os.path.join(self.conf.zopeDir(), 'bin','repozo.py' )
+class RepozoAction(ArgumentAction):
+    """Perform a variety of repozo backup/restore actions.
 
-        # make sure our software home is in the PYTHONPATH
-        env={};env['PYTHONPATH']="%s/lib/python" % self.conf.zopeDir(); os.environ.update(env)
-        command = ' '.join([python, repozo, arguments])
-        log.info("Backing up database file: %s to %s" % (source, target))
-        log.debug("Command used: %s" % command)
+    --backup and --restore are fine, but there are more diverse things
+    that can be done with repozo, this action allows some of them.
+    """
+
+    name = "Other backup/restore tasks with repozo."
+    option = '--repozo'
+    choices = ['sync',
+               'snapshot',
+               'restoresnapshot',
+               'full',
+               ]
+    
+    def addOption(self, optionparser):
+        optionparser.add_option(
+            self.option,
+            self.alternativeOption,
+            help=self.name,
+            action=self.kindOfAction,
+            default=False,
+            choices=self.choices)
+
+    def run(self, options=None):
+        choice = options.repozo
+        methodToRun = getattr(self, choice)
+        methodToRun(options)
+
+    def full(self, options=None):
+        # read paths from configuration
+        target = self.conf.backupBaseDir()
+        source = self.conf.databasePath()
+        actionutils.backup(conf=self.conf,
+                           sourceDatafs=source,
+                           targetDir=target,
+                           full=True)
+
+    def snapshot(self, options=None):
+        # read paths from configuration
+        target = self.conf.snapshotBaseDir()
+        source = self.conf.databasePath()
+        actionutils.backup(conf=self.conf,
+                           sourceDatafs=source,
+                           targetDir=target,
+                           full=True)
+
+    def restoresnapshot(self, options=None):
+        sourceDir = self.conf.snapshotBaseDir()
+        actionutils.restore(conf=self.conf,
+                            sourceDir=sourceDir)
+
+    def sync(self, options=None):
+        syncDatabase = self.conf.configData['sync_database']
+        if not syncDatabase:
+            log.warn("No sync_database specified, not syncing.")
+            return
+        log.debug("Syncing databases...")
+        # Create tempdir.
+        tempDir = tempfile.mkdtemp()
+        log.debug("Using tempdir %s.", tempDir)
+        # Make full backup of the source database.
+        source = syncDatabase
+        target = tempDir
+        actionutils.backup(conf=self.conf,
+                           sourceDatafs=source,
+                           targetDir=target,
+                           full=True)
+        # Restore that backup to the target.
+        sourceDir = tempDir
+        actionutils.restore(conf=self.conf,
+                            sourceDir=sourceDir)
+        # Zap the tempdir.
+        shutil.rmtree(tempDir)
 
-        os.system(command)
 
 class RestoreZopeAction(Action):
     """Restore the Zope Database of the instance.
@@ -510,70 +457,64 @@ class RestoreZopeAction(Action):
     backups from.
     """
 
-    name = "Restore the instance database from the last available backup."
-
-    def run(self):
-        # Lets make sure zope is stopped
-        action = StopZopeAction(self.conf)
-        action.run()
+    name = "Restore the database from the regular backup." 
+    option = '--restore'
 
+    def run(self, options=None):
         # read paths from configuration
-        backup_base_dir = self.conf.backupBaseDir()
-        source = os.path.join(backup_base_dir, self.conf.configData['project'])
-        datafs = self.conf.databasePath()
+        sourceDir = self.conf.backupBaseDir()
+        actionutils.restore(conf=self.conf,
+                            sourceDir=sourceDir)
 
-        # For now only supports full restores, no point in time option.
-        arguments = "--recover --output=%s --repository=%s" % (datafs, source)
 
-        # Now we have to remove the temp files, if they exist
-        for file in self.conf.databaseTempFiles():
-            if os.path.exists(file):
-                log.debug("Removing temporary database file: %s" % file)
-                os.remove(file)
+class RestoreDateZopeAction(ArgumentAction):
+    """Restore the Zope Database of the instance by date
 
-        python = self.conf.configData['python']
-        repozo = os.path.join(self.conf.zopeDir(), 'bin','repozo.py' )
+    This action relies on the backup_basedir_template variable to restore the
+    backups from.
+    """
 
-        # make sure our software home is in the PYTHONPATH
-        env={};env['PYTHONPATH']="%s/lib/python" % self.conf.zopeDir(); os.environ.update(env)
-        command = ' '.join([python, repozo, arguments])
-        log.info("Restoring database file: %s from backup repository: %s" % (datafs, source))
-        log.debug("Command used: %s" % command)
+    name = "Restore situation at DATE from regular backup." 
+    option = '--restore-date'
+    metavar = 'DATE'
 
-        os.system(command)
+    def run(self, options=None):
+        # read paths from configuration
+        sourceDir = self.conf.backupBaseDir()
+        fromTime = options.restore_date
+        actionutils.restore(conf=self.conf,
+                            sourceDir=sourceDir,
+                            fromTime=fromTime)
+
+
+def getActions():
+    """Return a list of action classes in this module.
+    """
+
+    moduleContents = globals()
+    actions = [moduleContents[key] for key in moduleContents.keys()
+               if key.endswith('Action')]
+    actions = [action for action in actions
+               if action.option]
+    def customSort(x, y):
+        return cmp(x.option, y.option)
+    actions.sort(customSort)
+    return actions
+
+def addOptions(optionparser):
+    for Action in getActions():
+        tempInstance = Action(None)
+        tempInstance.addOption(optionparser)
 
-# User-visible actions, this allows the user to do
-# 'instancemanager yourproject fresh', for instance.
-ACTIONS = {
-    'backup': BackupZopeAction,
-    'create': CreateInstanceAction,
-    'datafs': CopyDatafsAction,
-    'fg': ForegroundZopeAction,
-    'fresh': FreshInstallAction,
-    'printconfig': PrintConfigAction,
-    'products': AddProductsAction,
-    'restart': RestartZopeAction,
-    'restore': RestoreZopeAction,
-    'soft': SoftInstallAction,
-    'start': StartZopeAction,
-    'stop': StopZopeAction,
-    'zeostart': StartZeoAction,
-    'zeostop': StopZeoAction,
-    #'reinstall': QuickreinstallAction,
-    # Reinstall is not included as you need to stop zope first.
-    # It is best to call 'soft' or 'fresh'.
-    }
 
 if __name__ == '__main__':
     # Print out documentation
     print __doc__
     print
-    actionlist = list(ACTIONS.keys())
-    actionlist.sort()
-    for actionName in actionlist:
-        action = ACTIONS[actionName]
+    for action in getActions():
         doc = action.__doc__
         if not doc:
             doc = action.name
-        print '%s -- %s' % (actionName, doc)
+        print "'%s' -- %s" % (action.option, doc)
         print
+
